using System.Timers;
using GraphicsManager.Enums;
using GraphicsManager.Interfaces;
using GraphicsManager.Objects.Core;
using OpenTK.Graphics.OpenGL;
using OpenTK.Mathematics;
using OpenTK.Windowing.Common;
using OpenTK.Windowing.Common.Input;
using Timer = System.Timers.Timer;

namespace GraphicsManager.Objects;

public class FlowLayout : IRenderObject, IParent
{
    private Rectangle _bounds;

	public FlowLayout()
	{
		_bounds = new Rectangle();
		_bounds.Clicked += _bounds_Clicked;
		t = new(33);
		t.Enabled = true;
		t.Elapsed += TOnElapsed;
		t.Start();
		Controls.ControlAdded += ControlsOnControlAdded;
		Controls.ControlRemoved += ControlsOnControlRemoved;
		_bounds.MouseEnter += BoundsOnMouseEnter;
		_bounds.MouseLeave += BoundsOnMouseLeave;
	}

	public void ScrollToControl(IRenderObject Control)
	{
		if (Controls.Length < 1 || !Controls.Contains(Control)) throw new Exception("Control does not exist");
		if (Control.Location.Y >= 0 && (Control.Location.Y + Control.Size.Y) <= Size.Y) return;
		int index = Controls.IndexOf(Control);
		bool startbottom = false;
		if (Controls.Length - 1 - index < index) startbottom = true;
		int loc = 0;
		if ((Control.Location.Y + Control.Size.Y) >= Size.Y) loc = Size.Y - Control.Size.Y;
		int start = loc + Control.Size.Y;
		BlockDraw = true;
		if (startbottom)
		{
			for (int i = index + 1; i < Controls.Length; i++)
			{
				start += Controls[i].Size.Y;
			}

			for (int i = Controls.Length - 1; i >= 0; i--)
			{
				start -= Controls[i].Size.Y;
				Controls[i].Location = new(Controls[i].Location.X, start);
			}
		}
		else
		{
			for (int i = index; i >= 0; i--)
			{
				start -= Controls[i].Size.Y;
			}
			for (int i = 0; i < Controls.Length; i++)
			{
				start += Controls[i].Size.Y;
				Controls[i].Location = new(Controls[i].Location.X, start);
			}
		}

		for (int i = Controls.Length - 1; i >= 0; i--)
		{
			if (Controls[i].Location.Y + Controls[i].Size.Y < 0 || Controls[i].Location.Y > Size.Y)
			{
				Controls[i].Visible = false;
			}
		}
		BlockDraw = false;
		if (Parent is not null) Parent.TryDraw();
	}

	public void ScrollToTop()
	{
		if (Controls.Length < 1) return;
		ScrollToControl(Controls[0]);
	}

	public void ScrollToBottom()
	{
		if (Controls.Length < 1) return;
		ScrollToControl(Controls[Controls.Length - 1]);
	}
	
	public void ReportSizeUpdate(IRenderObject Control)
	{
		if (BlockDraw) return;
		BlockDraw = true;
		bool notfount = true;
		for (int i = 0; i < Controls.Length; i++)
		{
			if (Controls[i] != Control && notfount) continue;
			else
			{
				if (notfount)
				{
					notfount = false;
					continue;
				}

				Controls[i].Location = new(0, Controls[i - 1].Location.Y + Controls[i - 1].Size.Y);
			}
		}

		BlockDraw = false;
	}

	private Task ControlsOnControlRemoved()
	{
		if (Controls.Length < 1) return Task.CompletedTask;
		if (Controls[0].Location.Y > 0) Controls[0].Location = new(0, 0);
		for (int i = 1; i < Controls.Length; i++)
		{
			Controls[i].Location = new(0, Controls[Controls.Length - 1].Location.Y + Controls[Controls.Length - 1].Size.Y);
		}
		return Task.CompletedTask;
	}

	private Task ControlsOnControlAdded(IRenderObject arg)
	{
		if (Controls.Length > 0) arg.Location = new(0, Controls[Controls.Length - 1].Location.Y + Controls[Controls.Length - 1].Size.Y);
		else arg.Location = new(0, 0);
		if (arg is IParent par2)
		{
			for (int i = 0; i < par2.Controls.Length; i++)
			{
				par2.Controls[i].Distance = new(
					par2.Size.X - par2.Controls[i].Size.X - par2.Controls[i].Location.X,
					par2.Size.Y - par2.Controls[i].Size.Y - par2.Controls[i].Location.Y);
				
			}
		}
		arg.Size = new(Size.X, arg.Size.Y);
		if (arg is IParent par)
		{
			for (int i = 0; i < par.Controls.Length; i++)
			{
				bool top = (par.Controls[i].Anchor & ObjectAnchor.Top) == ObjectAnchor.Top;
				bool left = (par.Controls[i].Anchor & ObjectAnchor.Left) == ObjectAnchor.Left;
				bool right = (par.Controls[i].Anchor & ObjectAnchor.Right) == ObjectAnchor.Right;
				bool bottom = (par.Controls[i].Anchor & ObjectAnchor.Bottom) == ObjectAnchor.Bottom;
				if (!top && !bottom) { par.Controls[i].Anchor |= ObjectAnchor.Top; top = true; }
				if (!left && !right) { par.Controls[i].Anchor |= ObjectAnchor.Left; left = true; }
				int lx = (left ? par.Controls[i].Location.X : Size.X - par.Controls[i].Distance.X - par.Controls[i].Size.X);
				int ly = (top ? par.Controls[i].Location.Y : Size.Y - par.Controls[i].Distance.Y - par.Controls[i].Size.Y);
				int sy = (bottom ? Size.Y - par.Controls[i].Distance.Y - ly : par.Controls[i].Size.Y);
				int sx = (right ? Size.X - par.Controls[i].Distance.X - lx : par.Controls[i].Size.X);
				par.Controls[i].Size = new(sx, sy);
				par.Controls[i].Location = new(lx, ly);
			}
		}
		arg.Anchor = ObjectAnchor.Left | ObjectAnchor.Right | ObjectAnchor.Top;
		return Task.CompletedTask;
	}

	private Task _bounds_Clicked(IRenderObject arg)
	{
		if (Clicked is not null) _ = Clicked.Invoke(arg);
		return Task.CompletedTask;
	}
	private Task BoundsOnMouseLeave(IRenderObject arg)
	{
		inside = false;
		if (MouseLeave is not null) _ = MouseLeave.Invoke(this);
		return Task.CompletedTask;
	}
	
	public MouseCursor HoverMouse
	{
		get => _bounds.HoverMouse;
		set => _bounds.HoverMouse = value;
	}

	private bool inside = false;
	private Task BoundsOnMouseEnter(IRenderObject arg)
	{
		inside = true;
		if (MouseEnter is not null) _ = MouseEnter.Invoke(this);
		return Task.CompletedTask;
	}

	public ControlList Controls { get; } = new();
	public ObjectAnchor Anchor { get => _bounds.Anchor; set => _bounds.Anchor = value; }
	public Color4 BackgroundColor { get => _bounds.BackgroundColor; set => _bounds.BackgroundColor = value; }
	public bool Visible 
	{ 
		get => _bounds.Visible;
		set
		{
			BlockDraw = true;
			_bounds.Visible = value;
			if (!value)
				for (int i = 0; i < Controls.Length; i++)
					Controls[i].Visible = value;
			else
			{
				for (int i = 0; i < Controls.Length; i++)
					if (Controls[i].Location.Y > Size.Y || Controls[i].Location.Y + Controls[i].Size.Y < 0)
						Controls[i].Visible = false;
					else
						Controls[i].Visible = true;
			}
			if (Parent is not null) Parent.TryDraw();
			BlockDraw = false;
		} 
	}
	public Vector2i Size
	{
		get => _bounds.Size;
		set
		{
			BlockDraw = true;
			_bounds.Size = value; 
			for (int i = 0; i < Controls.Length; i++)
			{
				Controls[i].Size = new(value.X, Controls[i].Size.Y);
			}
			ParentResize(new());
			if (Parent is not null) Parent.TryDraw();
			BlockDraw = false;
		}
	}

	public Vector2 SizeAsFloat { get => _bounds.SizeAsFloat; }
	public void Focus()
	{
		
	}
	public void UnFocus()
	{
		
	}
	public Vector2i Location
	{
		get => _bounds.Location;
		set
		{
			BlockDraw = true;
			_bounds.Location = value;
			for (int i = 0; i < Controls.Length; i++)
			{
				Controls[i].Location = Controls[i].Location;
			}
			ParentResize(new());
			if (Parent is not null) Parent.TryDraw();
			BlockDraw = false;
		}
	}
	//public Vector2i ScissorLocation { get => _bounds.ScissorLocation; }

	public Vector2i Position => Location;
	public Vector2 LocationAsFloat { get => _bounds.LocationAsFloat; }
	public Vector2i Distance { get => _bounds.Distance; set => _bounds.Distance = value; }
	public event Func<IRenderObject, Task>? Clicked;
	public event Action<MouseButtonEventArgs> MouseDown;
	public event Action<KeyboardKeyEventArgs> KeyDown;
	public event Func<IRenderObject, Task>? WindowLoaded;
	public event Func<IRenderObject, Task>? MouseEnter;
	public event Func<IRenderObject, Task>? MouseLeave;
	public object? Tag { get; set; } = null;
	public IParent? Parent { get; private set; }
	public Window? Window { get; private set; }
	public bool Loaded { get; private set; } = false;

	public Vector2 MousePosition => Window!.MousePosition;

	public void LoadToParent(IParent Parent, Window Window)
	{
		if (Loaded) return;
		this.Parent = Parent;
		this.Window = Window;
		BlockDraw = true;
		Loaded = true;
		_bounds.LoadToParent(Parent, Window);
		for (int i = 0; i < Controls.Length; i++)
		{
			Controls[i].LoadToParent(this, Window);
		}
		BlockDraw = false;
		Window.MouseWheel += WindowOnMouseWheel;
		if (WindowLoaded is not null) WindowLoaded.Invoke(this);
	}
	
	public void ForceDistanceUpdate()
	{
		if (Parent is null) return;
		ForceDistanceUpdate(Parent);
	}
	
	public void ForceDistanceUpdate(IParent parent)
	{
		Distance = new(parent.Size.X - Size.X - Location.X, parent.Size.Y - Size.Y - Location.Y);
	}

	public bool BlockDraw { get; set; } = false;
	Timer t;
	private Queue<Action> scrols = new();
	private int dist = 0;
	public ContextMenu? ContextMenu { get => _bounds.ContextMenu; set => _bounds.ContextMenu = value; }
	
	public void TryDraw()
	{
		if (!BlockDraw) Parent!.TryDraw();
	}
	private void WindowOnMouseWheel(MouseWheelEventArgs obj)
	{
		try
		{
			if (!inside) return;
			BlockDraw = true;
			dist += (int)obj.OffsetY;
			if (scrols.Any())
			{
				for (int i = 0; i < scrols.Count; i++)
				{
					scrols.Dequeue().Invoke();
				}
				
			}
		}
		catch (Exception e)
		{
			Console.WriteLine(e);
		}
		
	}

	public uint HScrollPixels { get; set; } = 30;
	private void TOnElapsed(object? sender, ElapsedEventArgs e)
	{
		try
		{
			if (BlockDraw)
			{
				scrols.Enqueue(new Action(() =>
				{
					BlockDraw = true;
					int current = dist;
					dist = 0;
					if (Controls.Length < 1) return;
					bool down = current < 0;//scrole wheel dir
					bool moved = false;
					float totoaldist = (current * HScrollPixels);
					if (down && (!Controls[Controls.Length-1].Visible || (Controls[Controls.Length-1].Visible && Controls[Controls.Length-1].Location.Y + Controls[Controls.Length-1].Size.Y > Size.Y))) //put chat up
					{
						moved = true;
						int found = -1;
						bool lfound = false;
						for (int i = 0; i < Controls.Length; i++)
						{
							if (found == -1)
							{
								if (!Controls[i].Visible) continue;
								else
								{
									if ((Controls[i].Location.Y + Controls[i].Size.Y) + totoaldist < 0)
									{
										BlockDraw = true;
										Controls[i].Visible = false;
										continue;
									}

									found = Controls[i].Location.Y + (int)totoaldist;
								}
							}

							if (!lfound)
							{
								BlockDraw = true;
								Controls[i].Location = new(0, found);
								found += Controls[i].Size.Y;

								if (Controls[i].Location.Y <= Size.Y) Controls[i].Visible = true;
								else
								{
									lfound = true;
									Controls[i].Visible = false;
								}
								continue;
							}
							BlockDraw = true;
							if (Controls[i].Visible) Controls[i].Visible = false;
						}
					}
					if (!down && (!Controls[0].Visible || (Controls[0].Visible && Controls[0].Location.Y < 0))) // put chat down
					{
						moved = true;
						int found = -1;
						int lfound = -1;
						for (int i = Controls.Length - 1; i >= 0; i--)
						{
							if (found == -1)
							{
								if (!Controls[i].Visible) continue;
								else
								{
									if (Controls[i].Location.Y + totoaldist > Size.Y)
									{
										BlockDraw = true;
										Controls[i].Visible = false;
										continue;
									}

									found = Controls[i].Location.Y + (int)totoaldist + Controls[i].Size.Y;
								}
							}

							if (lfound == -1)
							{
								BlockDraw = true;
								found -= Controls[i].Size.Y;
								Controls[i].Location = new(0, found);

								if (Controls[i].Location.Y + Controls[i].Size.Y >= 0) Controls[i].Visible = true;
								else
								{
									lfound = i;
									Controls[i].Visible = false;
								}
								continue;
							}
							BlockDraw = true;
							if (Controls[i].Visible) Controls[i].Visible = false;
						}
					}

					if (moved)
					{
						Parent!.TryDraw();
					}

				}));
			}

			BlockDraw = false;
		}
		catch (Exception exception)
		{
			Console.WriteLine(exception);
		}
	}

	public bool dw;

	public void Draw(int x, int y, int w, int h)
	{
		if (Loaded)
		{
			int nx = x, ny = y, nw = w, nh = h;


			if (Location.X > nw)
				return;
			else
			{
				nx += Location.X;
				nw -= Location.X;
				if (Size.X < nw)
					nw = Size.X;
			}

			if (Location.Y > nh)
				return;
			else
			{
				ny += Location.Y;
				nh -= Location.Y;
				if (Size.Y < nh)
					nh = Size.Y;
			}
			if (dw) Console.WriteLine("Flowlayout\nLoc: {0}\nSize: {1}\n\nX: {2}\nY: {3}\nW: {4}\nH:{5}\n\nX: {6}\nY: {7}\nW: {8}\nH:{9}",
				Location,
				Size,
				x,
				y,
				w,
				h,
				nx,
				ny,
				nw,
				nh);
			if (nw ==0 || nh == 0) return;
			GL.Scissor(nx, Window!.Size.Y - ny - nh, nw, nh);
			_bounds.Draw(nx,ny,nw,nh);
			IEnumerable<IRenderObject> needload = Controls.Where(a => a.Loaded == false);

			if (needload.Any())
			{
				BlockDraw = true;
				foreach (IRenderObject Control in needload)
				{
					Control.LoadToParent(this, Window!);
				}

				BlockDraw = false;
			}
			for (int i = 0; i < Controls.Length; i++)
			{
				if (Controls[i] is not IParent)
				{
					GL.Scissor(nx, Window!.Size.Y - ny - nh, nw, nh);
				}
				Controls[i].Draw(nx, ny, nw, nh);
			}
		}
	}
    
	public void Clean()
	{
		for (int i = 0; i < Controls.Length; i++)
		{
			Controls[i].Clean();
		}
		_bounds.Clean();
	}

	public void ParentResize(ResizeEventArgs e)
	{
		if (e.Width == 0 && e.Height == 0) return;
		BlockDraw = true;
		for (int i = 0; i < Controls.Length; i++)
		{
			if (!Controls[i].Loaded) continue;
			bool top = (Controls[i].Anchor & ObjectAnchor.Top) == ObjectAnchor.Top;
			bool left = (Controls[i].Anchor & ObjectAnchor.Left) == ObjectAnchor.Left;
			bool right = (Controls[i].Anchor & ObjectAnchor.Right) == ObjectAnchor.Right;
			bool bottom = (Controls[i].Anchor & ObjectAnchor.Bottom) == ObjectAnchor.Bottom;
			if (!top && !bottom) { Controls[i].Anchor |= ObjectAnchor.Top; top = true; }
			if (!left && !right) { Controls[i].Anchor |= ObjectAnchor.Left; left = true; }
			int lx = (left ? Controls[i].Location.X : Size.X - Controls[i].Distance.X - Controls[i].Size.X);
			int ly = (top ? Controls[i].Location.Y : Size.Y - Controls[i].Distance.Y - Controls[i].Size.Y);
			int sy = (bottom ? Size.Y - Controls[i].Distance.Y - ly : Controls[i].Size.Y);
			int sx = (right ? Size.X - Controls[i].Distance.X - lx : Controls[i].Size.X);
			Controls[i].Size = new(sx, sy);
			Controls[i].Location = new(lx, ly);
			if (Controls[i] is IParent parent)
			{
				parent.ParentResize(e);
			}
		}
		Parent!.TryDraw();
		BlockDraw = false;
	}

	#region Cool Math Things
	public float[] RctToFloat(int x, int y, int Width, int Height, bool hastexture = false, float z = 0.0f)
	{
		if (hastexture)
		{
			return new float[20] {
				IntToFloat(x + Width), IntToFloat(y, true), z, 1.0f, 1.0f,// top r
                IntToFloat(x + Width), IntToFloat(y + Height, true), z, 1.0f, 0.0f,//b r
                IntToFloat(x), IntToFloat(y + Height, true), z, 0.0f, 0.0f,//bot l
                IntToFloat(x), IntToFloat(y, true), z, 0.0f, 1.0f// top l
            };
		}
		else
		{
			return new float[12] {
				IntToFloat(x + Width), IntToFloat(y, true), z,// top r
                IntToFloat(x + Width), IntToFloat(y + Height, true), z, //b r
                IntToFloat(x), IntToFloat(y + Height, true), z, //bot l
                IntToFloat(x), IntToFloat(y, true), z,// top l
            };
		}
	}

	public Vector3 PointToVector(float x, float y, float z = 0.0f)
	{
		return new Vector3(IntToFloat(x), IntToFloat(y, true), z);
	}

	public float IntToFloat(float p, bool Invert = false)
	{
		p += (Invert ? Location.Y : Location.X);
		IParent? tempp = Parent;
		while (tempp is not null)
		{
			p += (Invert ? tempp.Position.Y : tempp.Position.X);
			tempp = tempp.Parent;
		}
		float Size = (Invert ? Window!.Size.Y : Window!.Size.X);
		double half = Math.Round((double)Size / (double)2, 1);
		double Per = Math.Round((double)1 / half, 15);
		if (p == half) return 0.0f;
		if (Invert)
		{
			if (p > half) return (float)(((double)(p - half) * Per) * -1);
			else return (float)(1 - (p * Per));
		}
		else
		{
			if (p > half) return (float)((double)(p - half) * Per);
			else return (float)((1 - (p * Per)) * -1);
		}
	}

	public float FloatToInt(float p, bool Invert = false)
	{
		p += (Invert ? LocationAsFloat.Y : LocationAsFloat.X);
		IParent? tempp = Parent;
		while (tempp is not null)
		{
			p += (Invert ? tempp.LocationAsFloat.Y : tempp.LocationAsFloat.X);
			tempp = tempp.Parent;
		}

		float Size = (Invert ? Window!.Size.Y : Window!.Size.X);
		double half = Math.Round((double)Size / (double)2, 15);
		if (p == 0) return (int)half;
		if (Invert)
		{
			if (p < 0)
			{
				p *= -1;
				p++;
				return (float)(half * p);
			}
			else
			{
				return (float)(half - (p * half));
			}
		}
		else
		{
			if (p < 0)
			{
				p *= -1;
				p++;
				return (float)(Size - (half * p));
			}
			else
			{
				return (float)(p * half + half);
			}
		}
	}
	#endregion
}